"""
Unittests for `tcpnetlock.server` package.
"""
from tcpnetlock.server.action import Action


class TestActions:

    def test_valid_single_action(self):
        action = Action.from_line('lock1')
        assert action.name == 'lock1'

    def test_valid_action_with_one_param(self):
        action = Action.from_line('lock1,blocking:true')
        assert action.name == 'lock1'
        assert len(action.params) == 1
        assert action.params.get('blocking') == 'true'

    def test_valid_action_with_two_param(self):
        action = Action.from_line('lock1,blocking:true,foo:bar')
        assert action.name == 'lock1'
        assert len(action.params) == 2
        assert action.params.get('blocking') == 'true'
        assert action.params.get('foo') == 'bar'

    def test_valid_action_param_without_value(self):
        action = Action.from_line('lock1,blocking')
        assert action.name == 'lock1'
        assert len(action.params) == 1
        assert action.params.get('blocking') == ''

    def test_empty_action_fails(self):
        action = Action.from_line(',param:value')
        assert action
        assert not action.is_valid()

    def test_empty_param_fails(self):
        action = Action.from_line('action,:value')
        assert action
        assert not action.is_valid()



import logging
import socketserver

from tcpnetlock import constants as const
from tcpnetlock.common import ClientDisconnected
from tcpnetlock.protocol import Protocol
from tcpnetlock.server import action_handlers as handlers
from tcpnetlock.server.action import Action
from tcpnetlock.server.background_thread import BackgroundThread
from tcpnetlock.server.context import Context

"""
This implement a very simple network lock server based on just TCP.

The server listen for connections, reads the socket for a '\n' terminated line.
The received string is used as lock name.

If the lock is granted, it is valid until the client release it, or until the TCP connection is closed.

If the lock can't be granted (it's being held by another client), a response is sent and the TCP connection is closed.


The simplest way to use:

    1. SERVER accept TCP connection
    2. CLIENT send the lock name (utf-8 encoded), like: 'lock-name\n'
    3. SERVER tries to acquire the lock
        a. if lock was granted, returns 'ok\n' to the client and the TCP connections is KEPT OPEN
            - the LOCK remains granted until:
               + the client release it, sending 'release\n' to the server (and the TCP connection is CLOSED)
               + the client closes the TCP connection
               + for some reason the TCP connection is lost
        b. if lock was NOT granted, returns 'not-granted\n' to the client and the TCP connection is immediately CLOSED
"""

logger = logging.getLogger(__name__)


class TCPServer(socketserver.ThreadingTCPServer):
    allow_reuse_address = True
    daemon_threads = True
    DEFAULT_PORT = 7654

    def __init__(self, host='localhost', port=DEFAULT_PORT):
        super().__init__((host, port), TCPHandler)
        self._context = Context()
        self._background_thread = BackgroundThread(self._context)

    @property
    def port(self):
        return self.socket.getsockname()[1]

    def serve_forever(self, *args, **kwargs):
        self._background_thread.start()
        super().serve_forever(*args, **kwargs)


class TCPHandler(socketserver.BaseRequestHandler):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _invalid_request(self, protocol: Protocol, line: str):
        logger.warning("Received invalid request: '%s'", line)
        protocol.send(const.RESPONSE_INVALID_REQUEST)
        self.request.close()

    @property
    def _context(self) -> Context:
        return self.server._context

    def handle(self):
        self._context.requests_count.incr()
        protocol = Protocol(self.request)
        try:
            line = protocol.readline()
        except ClientDisconnected:
            logger.info("Client disconnected before getting line.")
            protocol.close()
            return

        action = Action.from_line(line)
        if action.is_valid():
            logger.debug("Received valid action: '%s'", action)
        else:
            return self._invalid_request(protocol, line)

        if action.name == const.ACTION_SERVER_SHUTDOWN:
            return handlers.ShutdownActionHandler(protocol, action, server=self.server).handle_action()

        if action.name == const.ACTION_PING:
            return handlers.PingActionHandler(protocol, action).handle_action()

        if action.name == const.ACTION_STATS:
            return handlers.StatsActionHandler(protocol, action, context=self._context).handle_action()

        if action.name != const.ACTION_LOCK:
            return handlers.InvalidActionActionHandler(protocol, action).handle_action()

        lock_name = action.params.get('name')
        if not const.VALID_LOCK_NAME_RE.match(lock_name):
            return handlers.InvalidLockActionHandler(protocol, action, lock_name=lock_name).handle_action()

        # Get the Lock, only while holding the GLOBAL LOCK
        # This is done to avoid 2 concurrent clients creating 2 instances of the same lock at the same time
        with self._context.global_lock:
            lock = self._context.locks[lock_name]

        # Acquire lock and proceed, or return failure.
        # If multiple concurrent clients try to get the lock, only one will proceed
        if lock.acquire_non_blocking():
            self._context.lock_acquired_count.incr()
            try:
                return handlers.LockGrantedActionHandler(
                    protocol, action, lock=lock, lock_name=lock_name).handle_action()
            finally:
                lock.release()
        else:
            self._context.lock_not_acquired_count.incr()
            return handlers.LockNotGrantedActionHandler(protocol, action, lock_name=lock_name).handle_action()



import collections
import threading
import typing

from tcpnetlock.common import Counter
from tcpnetlock.server.lock import Lock


class Context:

    def __init__(self):
        self._global_lock = threading.Lock()
        """Global lock to serialize modification of LOCKS"""

        self._locks = collections.defaultdict(Lock)
        """This dict contains the Lock instances"""

        self._requests_count = Counter()
        """How many requests were accepted"""

        self._lock_acquired_count = Counter()
        """How many times a lock was acquired"""

        self._lock_not_acquired_count = Counter()
        """How many times a lock was NOT acquired"""

    @property
    def global_lock(self) -> threading.Lock:
        return self._global_lock

    @property
    def locks(self) -> typing.Dict[str, Lock]:
        return self._locks

    @property
    def requests_count(self) -> Counter:
        return self._requests_count

    @property
    def lock_acquired_count(self) -> Counter:
        return self._lock_acquired_count

    @property
    def lock_not_acquired_count(self) -> Counter:
        return self._lock_not_acquired_count

    def counters(self):
        return {
            'requests_count': self._requests_count.count,
            'lock_acquired_count': self._lock_acquired_count.count,
            'lock_not_acquired_count': self._lock_not_acquired_count.count,
        }



import logging

from tcpnetlock.protocol import Protocol

logger = logging.getLogger(__name__)


class ClientAction:

    GENERATE_CUSTOM_MESSAGE = False

    def __init__(self, protocol: Protocol, message: str, valid_responses: list):
        self.protocol = protocol
        self.message = message
        self.valid_responses = valid_responses

        assert self.protocol
        assert self.message or self.GENERATE_CUSTOM_MESSAGE
        assert len(self.valid_responses)

    def parse_and_validate_response(self, line: str):
        response_code, *ignored = line.split(",", maxsplit=1)
        assert response_code in self.valid_responses,\
            "Invalid response: '{response_code}'. Valid responses: {valid_response_codes}. Full line: {line}".format(
                response_code=response_code,
                valid_response_codes=self.valid_responses,
                line=line
            )
        return response_code

    def read_valid_response(self):
        line = self.protocol.readline()
        response_code = self.parse_and_validate_response(line)
        return response_code

    def get_message(self, **kwargs):
        return self.message

    def handle(self, **kwargs):
        message = self.get_message(**kwargs)
        logger.debug("Sending message to server: %s", message)
        self.protocol.send(message)
        return self.read_valid_response()


class AcquireLockClientAction(ClientAction):

    GENERATE_CUSTOM_MESSAGE = True

    def get_message(self, lock_name, **kwargs):
        client_id = kwargs.pop('client_id')
        message = "lock,name:{lock_name}".format(lock_name=lock_name)
        if client_id:
            message += ",client-id:{client_id}".format(client_id=client_id)
        return message



"""
Tests for `tcpnetlock.client` and `tcpnetlock.server` packages.
"""
import json
import uuid

import pytest

from tcpnetlock import common
from tcpnetlock import constants
from tcpnetlock.client.client import LockClient
from .test_utils import BaseTest
from .test_utils import ServerThread
from .test_utils import lock_name
from .test_utils import lock_server

assert lock_name
assert lock_server


class TestLock(BaseTest):

    def test_server_is_alive(self, lock_server: ServerThread):
        """Test the test server is actually running"""
        assert lock_server.is_alive()
        client = lock_server.get_client()
        client.connect()
        client.ping()

    def test_connect_and_get_lock_works(self, lock_server):
        """Test that a lock can be acquired"""
        client = lock_server.get_client()
        client.connect()
        acquired = client.lock(uuid.uuid4().hex)
        assert acquired
        assert client.acquired
        client.close()

    def test_connect_and_get_lock_with_client_id_works(self, lock_server):
        """Test that a lock can be acquired"""
        name = uuid.uuid4().hex
        client_id = uuid.uuid4().hex

        client = lock_server.get_client(client_id=client_id)
        client.connect()
        acquired = client.lock(name)
        assert acquired
        client.close()

    def test_get_two_different_lock(self, lock_server):
        """Test that two different locks can be acquired"""
        client_1 = lock_server.get_client()
        client_1.connect()
        acquired = client_1.lock(uuid.uuid4().hex)
        assert acquired
        client_1.close()

        client_2 = lock_server.get_client()
        client_2.connect()
        acquired = client_2.lock(uuid.uuid4().hex)
        assert acquired
        client_2.close()

    def test_lock_twice_fails(self, lock_server):
        """Test that a locks can NOT be acquired twice"""
        name = uuid.uuid4().hex

        client_1 = lock_server.get_client()
        client_1.connect()
        acquired = client_1.lock(name)
        assert acquired

        client_2 = lock_server.get_client()
        client_2.connect()
        acquired = client_2.lock(name)
        assert not acquired
        assert not client_2.acquired

        client_1.close()
        client_2.close()

    def test_release_lock_and_re_acquire(self, lock_server):
        """Test that a locks can be acquired again after it's released"""
        name = uuid.uuid4().hex

        # acquire lock
        client_1 = lock_server.get_client()
        client_1.connect()
        acquired = client_1.lock(name)
        assert acquired
        # release lock
        client_1.release()
        client_1.close()

        lucky_client = self.get_client_with_lock_acquired(lock_server, name)
        assert lucky_client  # We got a client that re-acquired the lock

    def test_lock_is_released_when_client_closes_connection(self, lock_server):
        """Test that the server releases the lock if the socket connection is suddenly closed"""
        name = uuid.uuid4().hex

        # acquire lock
        client_1 = lock_server.get_client()
        client_1.connect()
        acquired = client_1.lock(name)
        assert acquired
        # don't release socket, just close the connection
        client_1.close()

        # re-acquire same lock
        lucky_client = self.get_client_with_lock_acquired(lock_server, name)
        assert lucky_client

    def test_server_rejects_invalid_lock_name(self, lock_server, monkeypatch):
        """Test that server do not accept invalid lock name"""
        invalid_names = (
            '.starts-with-space',
            'contains space',
            'contains%invalid%chars',
        )

        for invalid in invalid_names:
            client = lock_server.get_client()
            client.connect()

            with monkeypatch.context() as mp:
                mp.setattr(common.Utils, 'valid_lock_name', lambda x: True)
                acquired = client.lock(invalid)

            assert not acquired, "Lock granted for invalid lock name: '{invalid}'".format(invalid=invalid)
            client.close()

    def test_server_accept_valid_lock_name(self, lock_server, monkeypatch):
        """Test that server DO accept valid lock name"""
        valid_names = (
            'valid1--name',
            'valid2-name-8',
            'valid3_name',
            ' space4_is_ignored',
            'space5_is_ignored ',
            ' space6_is_ignored ',
        )

        for valid in valid_names:
            client = lock_server.get_client()
            client.connect()

            with monkeypatch.context() as mp:
                mp.setattr(common.Utils, 'valid_lock_name', lambda x: True)
                acquired = client.lock(valid)

            assert acquired, "Lock NOT granted for valid lock name: '{valid}'".format(valid=valid)
            client.close()

    def test_client_fails_with_invalid_lock_name(self, lock_server):
        """Test that client fails with invalid lock name"""
        invalid_names = (
            '.starts-with-space',
            'contains space',
            'contains%invalid%chars',
            'contains.point',
            'contains.point@and.arroba',
        )

        client = lock_server.get_client()
        client.connect()
        for invalid in invalid_names:
            with pytest.raises(common.InvalidLockNameError):
                client.lock(invalid)
        client.close()

    def test_client_fails_with_invalid_client_id(self):
        """Test that client fails with invalid client id"""
        invalid_client_ids = (
            'contains space',
            'contains%invalid%chars',
        )

        for invalid in invalid_client_ids:
            with pytest.raises(common.InvalidClientIdError):
                LockClient(client_id=invalid)


class TestWithLockGranted(BaseTest):

    def test_server_accept_keepalives(self, lock_server):
        """Test that server accept keepalives"""
        client = lock_server.get_client()
        client.connect()
        acquired = client.lock(uuid.uuid4().hex)
        assert acquired
        for _ in range(2):
            client.keepalive()
        client.close()

    def test_server_reports_invalid_action(self, lock_server):
        """Test that server reports an invalid action was sent while holding the lock"""
        client = lock_server.get_client()
        client.connect()
        acquired = client.lock(uuid.uuid4().hex)
        assert acquired
        client._protocol.send('invalid:action')
        line = client._protocol.readline(0.5)
        assert line == constants.RESPONSE_INVALID_ACTION

        # even after en invalid action, keep alive should continue working
        client.keepalive()
        client.close()


class TestAction(BaseTest):

    def test_server_rejects_invalid_action(self, lock_server):
        """Test that server report invalid action"""
        client = lock_server.get_client()
        client.connect()
        client._protocol.send('invalid:action')
        line = client._protocol.readline()
        assert line == constants.RESPONSE_INVALID_ACTION

    def test_server_rejects_invalid_request(self, lock_server):
        """Test that server report invalid request"""
        client = lock_server.get_client()
        client.connect()
        client._protocol.send(',param:value')
        line = client._protocol.readline()
        assert line == "bad-request"


class TestProtocol(BaseTest):

    def test_check_connection_with_valid_connection(self, lock_server: ServerThread, lock_name):
        client = lock_server.get_client()
        client.connect()
        assert client.lock(lock_name)

        # Should do nothing
        client.check_connection()

    # def test_check_connection_raises_with_invalid_connection(self, lock_server: ServerThread, lock_name):
    #     client = lock_server.get_client()
    #     client.connect()
    #     assert client.lock(lock_name)
    #
    #     # Should do nothing
    #     client.check_connection()
    #
    #     shutdown_client = lock_server.get_client()
    #     shutdown_client.connect()
    #     shutdown_client.server_shutdown()
    #
    #     # --
    #     lock_server._dont_shutdown = True  # hack? :/
    #     lock_server.join()
    #
    #     with pytest.raises(ClientDisconnected):
    #         client.check_connection()


class TestStatus(BaseTest):

    def test_server_serves_stats(self, lock_server):
        client = lock_server.get_client()
        client.connect()
        client._protocol.send('.stats')
        line = client._protocol.readline()
        stats = json.loads(line)
        assert 'lock_count' in stats
        assert 'maxrss' in stats



import socket
import threading
import time
import uuid

import pytest

from tcpnetlock.client.client import LockClient
from tcpnetlock.server.server import TCPServer


class BaseTest:

    def get_client_with_lock_acquired(self, lock_server, name, repeat=20, wait=0.3) -> LockClient:
        """
        Loop until we can get the lock. This is needed because the way the server works when RELEASING a lock:
        it first respond 'ok', and then it release the real lock... If a test case release the lock and then tries
        to re-ackquire it, the server might fail, so in this case we should retry a few more times.
        """
        iters = list(range(repeat))
        while iters:
            iters.pop()
            client = lock_server.get_client()
            client.connect()
            acquired = client.lock(name)
            if acquired:
                return client
            client.close()
            if iters:
                time.sleep(wait)

        return None


class ServerThread(threading.Thread):
    """
    Thread impl to run the server in a subthread
    """
    def __init__(self, port=0):
        super().__init__(daemon=True)
        self.server = TCPServer("localhost", port)
        self.port = self.server.port

    def get_client(self, **kwargs) -> LockClient:
        """Returns a client instance to connect to test server"""
        host = kwargs.pop('host', 'localhost')
        port = kwargs.pop('port', self.port)
        return LockClient(host, port, **kwargs)

    def run(self):
        self.server.serve_forever()


@pytest.fixture(scope='module')
def lock_server() -> ServerThread:
    """
    Fixture, returns the server process running a TCPServer ready to use
    """

    server_thread = ServerThread()
    server_thread.start()

    yield server_thread

    client = server_thread.get_client()
    client.connect()
    client.server_shutdown()
    client.close()

    server_thread.join(1)
    assert not server_thread.is_alive(), "Server didn't shut down cleanly"


@pytest.fixture()
def lock_name() -> str:
    return str(uuid.uuid4())


@pytest.fixture()
def free_tcp_port() -> int:
    """Returns an unused tcp port"""
    host = 'localhost'
    for port in range(65000, 65100):
        ssocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            ssocket.bind((host, port))
            return port
        except OSError as err:
            if err.errno == 98:
                continue
            raise  # unexpected error, re-raise it
        finally:
            ssocket.close()
    raise Exception("Couldn't find a free tcp port")



import logging
import socket

from tcpnetlock.constants import NEW_LINE_BINARY
from tcpnetlock.common import ClientDisconnected

logger = logging.getLogger(__name__)


class Protocol:
    """
    Encapsulates the read/write operations on the socket for the string, line-oriented protocol used
    by the server and client.
    """
    # FIXME: control length of line

    def __init__(self, sock: socket.socket):
        self.socket = sock
        self.buffer = bytearray()

    def close(self):
        logger.debug("socket.close()")
        self.socket.close()

    def send(self, message):
        logger.debug("socket.sendall('%s')", message)
        self.socket.sendall((message + '\n').encode())

    def _line_in_buffer(self):
        return self.buffer.find(NEW_LINE_BINARY) >= 0

    def _get_line_from_buffer(self) -> str:
        assert self._line_in_buffer()
        head, tail = self.buffer.split(NEW_LINE_BINARY, 1)
        self.buffer = tail
        return head.decode()

    def _readline_blocking(self) -> str:
        """Returns line, or raises ClientDisconnected if socket is closed"""
        logger.debug('Disabling socket timeout (will block)')
        self.socket.settimeout(None)
        while True:
            logger.debug('Reading from socket')
            recv_data = self.socket.recv(128)
            if not recv_data:
                raise ClientDisconnected()
            self.buffer.extend(recv_data)
            if self._line_in_buffer():
                return self._get_line_from_buffer()

    def _readline_non_blocking(self, timeout: int) -> str:
        """Wait for up to `timeout` for a line"""
        # FIXME: this implementation will wait AT LEAST `timeout`, but maybe more than that
        logger.debug('Setting socket timeout to %s', timeout)
        self.socket.settimeout(timeout)
        while True:
            if self._line_in_buffer():
                return self._get_line_from_buffer()
            try:
                logger.debug('Reading from socket')
                recv_data = self.socket.recv(128)
                if recv_data:
                    self.buffer.extend(recv_data)
                else:
                    raise ClientDisconnected()
            except socket.timeout:
                logger.debug('Reading from socket TIMED OUT')
                return None

    def readline(self, timeout=None) -> str:
        """
        Reads socket and returns a line.

        If partial data is received, but not an entire line, then None is returned.

        :return: line or None
        """

        if self._line_in_buffer():
            return self._get_line_from_buffer()

        if timeout is None:
            return self._readline_blocking()
        else:
            return self._readline_non_blocking(timeout=timeout)

    def check_connection(self):
        """
        Raises ClientDisconnected if connection is closed (this is used to detect cases like if the server died).
        """
        # FIXME: must be a better way to implement this check :/
        self.socket.settimeout(1)
        try:
            recv_data = self.socket.recv(1)
        except socket.timeout:
            return  # this is not a guaranty that TCP connection still exists

        if recv_data:
            self.buffer.extend(recv_data)
        else:
            raise ClientDisconnected()



"""
Unittests for `tcpnetlock.protocol` package.
"""

import socket
from unittest import mock

import pytest

from tcpnetlock.common import ClientDisconnected
from tcpnetlock.protocol import Protocol


class TestProtocol:

    def test_non_blocking_protocol(self):
        mock_socket = mock.Mock()
        mock_socket.recv = mock.MagicMock(side_effect=[socket.timeout()])
        protocol = Protocol(mock_socket)
        line = protocol.readline(0.1)
        assert line is None

    def test_reading_full_line_non_blocking(self):
        mock_socket = mock.Mock()
        mock_socket.recv = mock.MagicMock(side_effect=["full-line\n".encode()])
        protocol = Protocol(mock_socket)
        line = protocol.readline(0.1)
        assert line == 'full-line'

    def test_reading_full_line_blocking(self):
        mock_socket = mock.Mock()
        mock_socket.recv = mock.MagicMock(side_effect=["full-line\n".encode()])
        protocol = Protocol(mock_socket)
        line = protocol.readline()
        assert line == 'full-line'

    def test_reading_partial_line_blocking(self):
        mock_socket = mock.Mock()
        mock_socket.recv = mock.MagicMock(side_effect=["partial-".encode(),
                                                       "line\n".encode()])
        protocol = Protocol(mock_socket)
        line = protocol.readline()
        assert line == 'partial-line'

    def test_reading_partial_line_non_blocking(self):
        mock_socket = mock.Mock()
        mock_socket.recv = mock.MagicMock(side_effect=["partial-".encode(),
                                                       "line\n".encode()])
        protocol = Protocol(mock_socket)
        line = protocol.readline(0.1)
        assert line == 'partial-line'

    def test_reading_after_receiving_multiple_lines_blocking(self):
        mock_socket = mock.Mock()
        mock_socket.recv = mock.MagicMock(side_effect=["line1\nline2\nline3\n".encode()])
        protocol = Protocol(mock_socket)

        line = protocol.readline()
        assert line == 'line1'

        line = protocol.readline()
        assert line == 'line2'

        line = protocol.readline()
        assert line == 'line3'

    def test_reading_after_receiving_multiple_lines_non_blocking(self):
        mock_socket = mock.Mock()
        mock_socket.recv = mock.MagicMock(side_effect=["line1\nline2\nline3\n".encode()])
        protocol = Protocol(mock_socket)

        line = protocol.readline(0.1)
        assert line == 'line1'

        line = protocol.readline(0.1)
        assert line == 'line2'

        line = protocol.readline(0.1)
        assert line == 'line3'

    def test_disconnect_is_detected_non_blocking(self):
        mock_socket = mock.Mock()
        mock_socket.recv = mock.MagicMock(side_effect=["partial-".encode(),
                                                       ''.encode()])
        protocol = Protocol(mock_socket)

        with pytest.raises(ClientDisconnected):
            protocol.readline(0.1)

    def test_disconnect_is_detected_blocking(self):
        mock_socket = mock.Mock()
        mock_socket.recv = mock.MagicMock(side_effect=["partial-".encode(),
                                                       ''.encode()])
        protocol = Protocol(mock_socket)

        with pytest.raises(ClientDisconnected):
            protocol.readline()



import logging
import os
import queue
import subprocess
import sys
import threading
import time

from tcpnetlock import constants
from tcpnetlock.cli import common
from tcpnetlock.client import client
from tcpnetlock.server import server

logger = logging.getLogger(__name__)


ERR_INVALID_OPTIONS = 2  # like unix commands
ERR_LOCK_NOT_GRANTED = 123
ERR_EXECUTING_COMMAND = 124
ERR_CONNECTION_REFUSED = 125
ERR_FILE_NOT_FOUND = 127  # like bash


class Main(common.BaseMain):

    DEFAULT_HOST = os.environ.get('TCPNETLOCK_HOST', 'localhost')
    DEFAULT_PORT = os.environ.get('TCPNETLOCK_PORT', str(server.TCPServer.DEFAULT_PORT))
    DEFAULT_CLIENT_ID = os.environ.get('TCPNETLOCK_CLIENT_ID')

    def add_app_arguments(self):
        self.parser.description = """
        Connect to TCPNetLock server and tries to get the lock.
        If lock IS GRANTED, the client executes the command provided by the user.
        If the lock IS NOT GRANTED, we exits with status {not_granted}.

        The following environment variables are used: TCPNETLOCK_HOST, TCPNETLOCK_PORT and TCPNETLOCK_CLIENT_ID.
        """.format(not_granted=ERR_LOCK_NOT_GRANTED)

        parser = self.parser
        parser.add_argument("--lock-name",
                            help="Name of the lock to acquire")

        parser.add_argument("--host",
                            default=self.DEFAULT_HOST)

        parser.add_argument("--port",
                            default=self.DEFAULT_PORT,
                            type=common.PositiveInteger())

        parser.add_argument("--retry",
                            default=0,
                            type=common.PositiveInteger(),
                            help="How many times retry getting the lock (defauls 0, do not retry)")

        parser.add_argument("--retry-wait",
                            default=10,
                            type=common.PositiveInteger(),
                            help="How many seconds wait between tries")

        parser.add_argument("--client-id",
                            default=self.DEFAULT_CLIENT_ID,
                            help="Client id to report to the server")

        parser.add_argument("--keep-alive",
                            default=False,
                            action='store_true')

        parser.add_argument("--keep-alive-secs",
                            default=15,
                            type=common.PositiveInteger())

        parser.add_argument("--shell",
                            default=False,
                            action='store_true',
                            help="Invoke the command with a shell (useful if you're using piping or redirecting)")

        parser.add_argument("command",
                            nargs='+',
                            help="Command to execute (if lock is granted)")

    def validate_and_fix_parameters(self):
        assert self.args

        # --- Get 'command' to use (depending on value of '--shell')
        if self.args.shell:
            if len(self.args.command) != 1:
                self.parser.error("When invoking with --shell, you should provide a SINGLE command. "
                                  "If you're having problems to achieve that, try wrapping it with quotes.")
                sys.exit(ERR_INVALID_OPTIONS)

        # --- Get 'lock name' to use (provided by user, or generated from command)
        if not self.args.lock_name:
            temporary = ' '.join(self.args.command)
            temporary = temporary.replace(' ', '_')
            temporary = temporary.replace('.', '_')
            temporary = temporary.replace('/', '_')
            temporary = temporary.strip('_')
            matches = constants.VALID_CHARS_IN_LOCK_NAME_RE.findall(temporary)
            self.args.lock_name = ''.join(matches)

            logger.info("Generated lock name '%s' (from provided command '%s')", self.args.lock_name, self.args.command)

            if not self.args.lock_name:
                self.parser.error("Couldn't create a lock name from the command. "
                                  "You must specify the lock name with --lock-name")
                sys.exit(ERR_INVALID_OPTIONS)

    def start_keepalive_thread(self, lock_client: client.LockClient):
        keepalive_queue = queue.Queue()

        def loop_keepalives():
            while True:
                logger.debug("Sleeping for %s before sending keep-alive", self.args.keep_alive_secs)
                for _ in range(self.args.keep_alive_secs):
                    try:
                        msg = keepalive_queue.get(block=True, timeout=1)
                        if msg is None:
                            logger.info("Shutting down keepalive thread...")
                            return
                        else:
                            logger.warning("Invalid message '%s' received from Queue. This is rare.", msg)
                    except queue.Empty:
                        pass
                logger.info("Sending keepalive")
                lock_client.keepalive()

        keepalive_thread = threading.Thread(target=loop_keepalives, daemon=True)
        keepalive_thread.start()

        return keepalive_thread, keepalive_queue

    def stop_keepalive_thread(self, keepalive_thread, keepalive_queue):
        if keepalive_thread is None or keepalive_queue is None:
            return
        try:
            keepalive_queue.put(None)
            keepalive_thread.join(2)
            if keepalive_thread.is_alive():
                logger.warning("keepalive_thread still alive: %s", keepalive_thread)
        except:  # noqa: E722 we need to ignore any error
            logger.warning("Error occurred while stopping background thread")

    def main(self):
        self.validate_and_fix_parameters()

        tries = list(range(self.args.retry + 1))
        while tries:
            tries.pop()

            # --- Try to get lock
            lock_client = client.LockClient(self.args.host, self.args.port, client_id=self.args.client_id)
            try:
                lock_client.connect()
            except ConnectionRefusedError:
                logger.error("Connection refused. Server: '%s:%s'", self.args.host, self.args.port)
                sys.exit(ERR_CONNECTION_REFUSED)

            granted = lock_client.lock(self.args.lock_name)
            if not granted:
                if tries:
                    logger.info("Lock '%s' not granted. Still %s retries pending. Will retry in %s seconds...",
                                self.args.lock_name, len(tries), self.args.retry_wait)
                    time.sleep(self.args.retry_wait)
                else:
                    logger.info("Lock '%s' not granted. Exiting...", self.args.lock_name)
                    sys.exit(ERR_LOCK_NOT_GRANTED)

        # --- Send keepalive from thread
        if self.args.keep_alive:
            keepalive_thread, keepalive_queue = self.start_keepalive_thread(lock_client)
        else:
            keepalive_thread = keepalive_queue = None

        # --- Call command
        logger.info("Lock '%s' was granted. Proceeding with command '%s'", self.args.lock_name, self.args.command)
        completed_process = None
        try:
            completed_process = subprocess.run(self.args.command,
                                               shell=self.args.shell,
                                               stdout=None,
                                               stderr=None,
                                               stdin=None,
                                               check=False)
        except FileNotFoundError as err:
            sys.stderr.write("ERROR: command not found: '{command}'\n".format(command=err.filename))
            sys.exit(ERR_FILE_NOT_FOUND)
        except KeyboardInterrupt:
            pass

        # --- Cleanup
        self.stop_keepalive_thread(keepalive_thread, keepalive_queue)

        # Release AFTER keepalive thread is stopped.
        # Otherwise, the 2 threads could be concurrently reading/writing to the socket
        lock_client.release()

        # --- Exit with same 'exit status' of the process we have just ran
        if completed_process:
            sys.exit(completed_process.returncode)
        else:
            sys.exit(ERR_EXECUTING_COMMAND)  # something happened (including KeyboardInterrupt)


def main():
    Main().run(args=sys.argv[1:])


if __name__ == '__main__':
    main()



"""
Tests for `tcpnetlock.client` and `tcpnetlock.server` packages.
"""
import subprocess

from tcpnetlock.cli import tnl_server
from .test_utils import BaseTest
from .test_utils import ServerThread
from .test_utils import lock_server

assert lock_server


class TestServerCli(BaseTest):
    def test_report_bind_to_used_port(self, lock_server: ServerThread):
        args = [
            'python', '-m', 'tcpnetlock.cli.tnl_server',
            '--port={port}'.format(port=lock_server.port),
        ]
        completed_process = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        assert completed_process.returncode == tnl_server.ERR_SERVER_BIND
        assert completed_process.stderr.decode().find('[Errno 98]') >= 0

    def test_report_bind_to_privileged_port(self, lock_server: ServerThread):
        args = [
            'python', '-m', 'tcpnetlock.cli.tnl_server',
            '--port=20',
        ]
        completed_process = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        assert completed_process.returncode == tnl_server.ERR_SERVER_BIND
        assert completed_process.stderr.decode().find('[Errno 13]') >= 0



class Action:

    def __init__(self, action: str, params: dict):
        self.action = action
        self.params = params

    @property
    def name(self):
        return self.action

    def is_valid(self):
        # action, and param keys must be non-empty
        return \
            len(self.action) > 0 \
            and all(len(p) > 0 for p in self.params.keys())

    @classmethod
    def from_line(cls, line: str):
        action, *raw_params = line.split(',')
        params = []
        for key, *values in [p.split(':', 1) for p in raw_params]:
            if values:
                assert len(values) == 1
                param = [key.strip(), values[0].strip()]
            else:
                param = [key.strip(), '']
            params.append(param)
        return Action(action.strip(), dict(params))

    def __str__(self):
        if self.params:
            params = ', '.join(['{}:{}'.format(k, v)
                                for k, v in self.params.items()])
            return "Action: '{name}'; {params}".format(name=self.action, params=params)
        else:
            return "Action: '{name}'".format(name=self.action)



import logging
import sys

from tcpnetlock import __version__ as tcpnetlock_version
from tcpnetlock.server import server
from tcpnetlock.cli import common

logger = logging.getLogger(__name__)


ERR_SERVER_BIND = 2
ERR_HANDLING_REQUESTS = 3


class Main(common.BaseMain):

    def add_app_arguments(self):
        self.parser.add_argument("--listen", default='localhost')
        self.parser.add_argument("--port", default=server.TCPServer.DEFAULT_PORT, type=int)

    @property
    def version(self):
        return tcpnetlock_version

    def main(self):
        logger.info("Started server v%s listening on %s:%s", self.version, self.args.listen, self.args.port)
        try:
            lock_server = server.TCPServer(self.args.listen, self.args.port)
        except BaseException as err:
            logger.debug('Error while bind()ing...', exc_info=True)
            print(str(err) or 'Error detected while creating server', file=sys.stderr)
            sys.exit(ERR_SERVER_BIND)

        try:
            lock_server.serve_forever()
        except BaseException as err:
            logger.debug('Error while handling requests...', exc_info=True)
            print(str(err) or 'Error while handling requests', file=sys.stderr)
            sys.exit(ERR_HANDLING_REQUESTS)


def main():
    Main().run(args=sys.argv[1:])


if __name__ == '__main__':
    main()



import tcpnetlock.constants


# def ignore_client_disconnected_exception(f):
#     """Swallows ClientDisconnected. Use this in method that can receive this exception but can safely ignore it"""
#     @functools.wraps(f)
#     def wrapper(*args, **kwargs):
#         try:
#             return f(*args, **kwargs)
#         except ClientDisconnected:
#             pass
#     return wrapper


class TcpNetLockException(Exception):
    """
    Base class for exceptions.
    """


class ClientDisconnected(TcpNetLockException):
    pass


class InvalidClientIdError(TcpNetLockException):
    """
    Raised by the client if the provided client-id is not valid.
    """


class InvalidLockNameError(TcpNetLockException):
    """
    Raised by the client if the provided LOCK name is not valid.
    """


class Utils:

    @staticmethod
    def valid_lock_name(lock_name):
        """Returns True if the provided lock name is valid"""
        return bool(tcpnetlock.constants.VALID_LOCK_NAME_RE.match(lock_name))

    @staticmethod
    def validate_client_id(client_id, accept_none=True):
        """Raises InvalidClientIdError if client-id is invalid. Pass if it's None"""
        if client_id is None:
            if accept_none:
                return
            else:
                raise InvalidClientIdError("You must provide a client-id")

        if not tcpnetlock.constants.VALID_CLIENT_ID_RE.match(client_id):
            raise InvalidClientIdError("The provided client-id is not valid")


class Counter:
    """Counter, just that.
    We don't care about atomicity.
    """

    def __init__(self):
        self._count = 0

    @property
    def count(self):
        return self._count

    def incr(self):
        self._count += 1



import logging
import os
import sys
import time

from tcpnetlock.client import client
from tcpnetlock.cli import common
from tcpnetlock.common import ClientDisconnected

logger = logging.getLogger(__name__)


ERR_CONNECTION_REFUSED = 2
ERR_CONNECTION_FAILED = 3
ERR_UNKNOWN = 4

ERR_TCP_DISCONNECT_WHILE_HOLDING_LOCK = 122
ERR_LOCK_NOT_GRANTED = 123


class Main(common.BaseMain):

    DEFAULT_HOST = os.environ.get('TCPNETLOCK_HOST', 'localhost')
    DEFAULT_PORT = os.environ.get('TCPNETLOCK_PORT', str(client.LockClient.DEFAULT_PORT))
    DEFAULT_CLIENT_ID = os.environ.get('TCPNETLOCK_CLIENT_ID')

    def add_app_arguments(self):
        self.parser.description = """
        Connect to TCPNetLock server and tries to get the lock.
        If lock IS GRANTED, the client enters a busy loop, sending keepalives or checking
        that the TCP connection is alive.
        If the lock IS NOT GRANTED, we exits with status {not_granted}.

        The following environment variables are used: TCPNETLOCK_HOST, TCPNETLOCK_PORT and TCPNETLOCK_CLIENT_ID.
        """.format(not_granted=ERR_LOCK_NOT_GRANTED)

        self.parser.add_argument("lock_name")
        self.parser.add_argument("--host", default=self.DEFAULT_HOST)
        self.parser.add_argument("--port", default=self.DEFAULT_PORT, type=int)
        self.parser.add_argument("--client-id", default=self.DEFAULT_CLIENT_ID)
        self.parser.add_argument("--keep-alive", default=False, action='store_true')
        self.parser.add_argument("--keep-alive-secs", default=15, type=int)

    def main(self):
        lock_client = client.LockClient(self.args.host, self.args.port, client_id=self.args.client_id)

        try:
            lock_client.connect()
        except ConnectionRefusedError as err:
            logger.debug("ConnectionRefusedError", exc_info=True)
            print(str(err), file=sys.stderr)
            sys.exit(ERR_CONNECTION_REFUSED)
        except BaseException as err:
            logger.debug("Can't connect", exc_info=True)
            print(str(err), file=sys.stderr)
            sys.exit(ERR_CONNECTION_FAILED)

        try:
            granted = lock_client.lock(self.args.lock_name)
            if not granted:
                logger.debug("Lock '%s' not granted. Exiting...", self.args.lock_name)
                print("ERROR: lock '{lock}' not granted by server".format(lock=self.args.lock_name), file=sys.stderr)
                sys.exit(ERR_LOCK_NOT_GRANTED)

            if self.args.keep_alive:
                while True:
                    logger.info("Sleeping for %s... (after that, will send a keep-alive)", self.args.keep_alive_secs)
                    time.sleep(self.args.keep_alive_secs)
                    lock_client.keepalive()
            else:
                while True:
                    logger.debug("Checking connection...")
                    try:
                        lock_client.check_connection()
                    except ClientDisconnected as err:
                        logger.debug("Unexpected disconnection while holding the lock '%s'. "
                                     "Server was killed?", self.args.lock_name)
                        print("ERROR: Unexpected disconnection while holding the lock '{lock}'. "
                              "Server was killed?".format(lock=self.args.lock_name),
                              file=sys.stderr)
                        sys.exit(ERR_TCP_DISCONNECT_WHILE_HOLDING_LOCK)

                    time.sleep(5)

        except SystemExit as err:
            sys.exit(err.code)

        except BaseException as err:
            logger.debug("Error while getting or having the lock", exc_info=True)
            print(str(err), file=sys.stderr)
            sys.exit(ERR_UNKNOWN)
        finally:
            lock_client.close()


def main():
    Main().run(args=sys.argv[1:])


if __name__ == '__main__':
    main()



import logging
import socket

from tcpnetlock import common
from tcpnetlock import constants
from tcpnetlock.client.action import AcquireLockClientAction
from tcpnetlock.client.action import ClientAction
from tcpnetlock.common import Utils
from tcpnetlock.protocol import Protocol
from tcpnetlock.server import server

logger = logging.getLogger(__name__)


class LockClient:
    DEFAULT_PORT = server.TCPServer.DEFAULT_PORT

    def __init__(self, host='localhost', port=DEFAULT_PORT, client_id=None):
        """
        Creates a client to connect to the server.

        :param host: hostname of the server
        :param port: port to connect
        """
        self._host = host
        self._port = port
        self._acquired = None
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._protocol = Protocol(self._socket)
        self._client_id = client_id
        Utils.validate_client_id(client_id)

    def connect(self):
        """
        Connects to server.

        :raises ConnectionRefusedError if connection is refused
        :return:
        """
        logger.info("Connecting to '%s:%s'...", self._host, self._port)
        self._socket.connect((self._host, self._port))

    def lock(self, name: str) -> bool:
        """
        Tries to acquire a lock

        :param name: lock name
        :return: boolean indicating if lock as acquired or not
        """
        if not Utils.valid_lock_name(name):
            raise common.InvalidLockNameError("Lock name is invalid: '{lock_name}'".format(lock_name=name))

        response_code = AcquireLockClientAction(
            self._protocol,
            None,
            [constants.RESPONSE_OK,
             constants.RESPONSE_LOCK_NOT_GRANTED,
             constants.RESPONSE_ERR]
        ).handle(lock_name=name, client_id=self._client_id)

        # FIXME: raise specific exception if RESPONSE_ERR is received (ex: InvalidClientId)
        self._acquired = bool(response_code == constants.RESPONSE_OK)
        return self._acquired

    def server_shutdown(self):
        """Send order to shutdown the server"""
        return ClientAction(self._protocol,
                            constants.ACTION_SERVER_SHUTDOWN,
                            [constants.RESPONSE_SHUTTING_DOWN]).handle()

    def ping(self):
        """Send ping to the server"""
        return ClientAction(self._protocol,
                            constants.ACTION_PING,
                            [constants.RESPONSE_PONG]).handle()

    def keepalive(self):
        """Send a keepalive to the server"""
        return ClientAction(self._protocol,
                            constants.ACTION_KEEPALIVE,
                            [constants.RESPONSE_STILL_ALIVE]).handle()

    def release(self):
        """Release the held lock"""
        return ClientAction(self._protocol,
                            constants.ACTION_RELEASE,
                            [constants.RESPONSE_RELEASED]).handle()

    def check_connection(self):
        """
        Raises ClientDisconnected if connection is closed (this is used to detect cases like if the server died.
        """
        self._protocol.check_connection()

    def close(self):
        """Close the socket. As a result of the disconnection, the lock will be released at the server."""
        logger.debug("Closing the socket...")
        self._protocol.close()

    @property
    def acquired(self) -> bool:
        """
        Returns boolean indicating if lock was acquired or not
        :return: True if lock was acquired, False otherwise
        """
        assert self._acquired in (True, False)  # Fail if lock() wasn't called
        return self._acquired



import threading
import time


class Lock:

    def __init__(self):
        self._lock = threading.Lock()
        self._name = None
        self._timestamp = 0
        self._client_id = None

    def acquire_non_blocking(self):
        return self._lock.acquire(blocking=False)

    def update(self, lock_name, client_id):
        if self._name is None:
            self._name = lock_name
        else:
            assert self._name == lock_name

        self._timestamp = time.monotonic()
        self._client_id = client_id

    def release(self):
        self._lock.release()

    @property
    def locked(self):
        return self._lock.locked()

    @property
    def age(self):
        return time.monotonic() - self._timestamp

    def __str__(self):
        return "Lock: '{name}', client-id: '{client}', age: {age}".format(
            name=self._name, client=self._client_id or '', age=int(self.age))



import json
import logging
import resource

from tcpnetlock import constants as const
from tcpnetlock.common import ClientDisconnected
from tcpnetlock.protocol import Protocol
from tcpnetlock.server.action import Action
from tcpnetlock.server.context import Context

logger = logging.getLogger(__name__)


__all__ = [
    'ShutdownActionHandler',
    'PingActionHandler',
    'StatsActionHandler',
    'InvalidLockActionHandler',
    'LockNotGrantedActionHandler',
    'LockGrantedActionHandler',
    'InnerReleaseLockActionHandler',
    'InnerKeepAliveActionHandler',
]


class ActionHandler:
    def __init__(self, protocol: Protocol, action: Action):
        self.protocol = protocol
        self.action = action


class ShutdownActionHandler(ActionHandler):

    def __init__(self, *args, **kwargs):
        self.server = kwargs.pop('server')
        super().__init__(*args, **kwargs)

    def can_proceed(self):
        # FIXME: assert connections came from localhost
        return True

    def handle_action(self):
        if not self.can_proceed():
            pass  # FIXME: do something
        self.protocol.send(const.RESPONSE_SHUTTING_DOWN)
        self.server.shutdown()
        self.protocol.close()


class PingActionHandler(ActionHandler):

    def handle_action(self):
        self.protocol.send(const.RESPONSE_PONG)
        self.protocol.close()


class StatsActionHandler(ActionHandler):

    def __init__(self, *args, **kwargs):
        self._context = self.__pop_context(kwargs)
        super().__init__(*args, **kwargs)

    def __pop_context(self, kwargs) -> Context:
        return kwargs.pop('context')

    def _get_maxrss(self):
        try:
            return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        except:  # noqa: E722 we need to ignore any error
            logger.warning("resource.getrusage() failed", exc_info=True)
            return 'n/a'

    def handle_action(self):
        stats = {
            'lock_count': len(self._context.locks),
            'maxrss': self._get_maxrss(),
        }
        stats.update(self._context.counters())
        self.protocol.send(json.dumps(stats))
        self.protocol.close()


class InvalidActionActionHandler(ActionHandler):
    def handle_action(self):
        logger.warning("Received invalid action: '%s'", self.action.name)
        self.protocol.send(const.RESPONSE_INVALID_ACTION)
        self.protocol.close()


class InvalidLockActionHandler(ActionHandler):
    def __init__(self, *args, **kwargs):
        self.lock_name = kwargs.pop('lock_name')
        super().__init__(*args, **kwargs)

    def handle_action(self):
        logger.warning("Received invalid lock name: '%s'", self.lock_name)
        self.protocol.send(const.RESPONSE_ERR + ',invalid lock name')
        self.protocol.close()


class LockNotGrantedActionHandler(ActionHandler):
    def __init__(self, *args, **kwargs):
        self.lock_name = kwargs.pop('lock_name')
        super().__init__(*args, **kwargs)

    def handle_action(self):
        logger.info("Lock NOT granted: %s", self.lock_name)
        self.protocol.send(const.RESPONSE_LOCK_NOT_GRANTED)
        self.protocol.close()


class LockGrantedActionHandler(ActionHandler):

    def __init__(self, *args, **kwargs):
        self.lock = kwargs.pop('lock')
        self.lock_name = kwargs.pop('lock_name')
        super().__init__(*args, **kwargs)
        self.client_id = self.action.params.get('client-id')

    def handle_action(self):
        try:
            self._handle_action()
        except ClientDisconnected:
            logger.info("ClientDisconnected: lock will be released: %s", self.lock)

    def _handle_action(self):
        self.lock.update(self.lock_name, self.client_id)
        self.protocol.send(const.RESPONSE_OK)
        logger.info("Lock granted: %s", self.lock)

        while True:
            line = self.protocol.readline(timeout=1.0)
            if line is None:
                continue

            inner_action = Action.from_line(line)
            logger.debug("Inner action: '%s' for lock %s", inner_action, self.lock)

            # FIXME: handle 'invalid requests' here too

            if inner_action.name == const.ACTION_RELEASE:
                return InnerReleaseLockActionHandler(self.protocol, inner_action, lock=self.lock).handle_action()
            if inner_action.name == const.ACTION_KEEPALIVE:
                InnerKeepAliveActionHandler(self.protocol, inner_action, lock=self.lock).handle_action()
            else:
                InnerInvalidActionActionHandler(self.protocol, inner_action, lock=self.lock).handle_action()


class InnerReleaseLockActionHandler(ActionHandler):

    def __init__(self, *args, **kwargs):
        self.lock = kwargs.pop('lock')
        super().__init__(*args, **kwargs)

    def handle_action(self):
        logger.info("Releasing lock: %s", self.lock)
        self.protocol.send(const.RESPONSE_RELEASED)
        self.protocol.close()


class InnerKeepAliveActionHandler(ActionHandler):

    def __init__(self, *args, **kwargs):
        self.lock = kwargs.pop('lock')
        super().__init__(*args, **kwargs)

    def handle_action(self):
        logger.debug("Received keepalive from client. Lock: %s", self.lock)
        self.protocol.send(const.RESPONSE_STILL_ALIVE)


class InnerInvalidActionActionHandler(ActionHandler):

    def __init__(self, *args, **kwargs):
        self.lock = kwargs.pop('lock')
        super().__init__(*args, **kwargs)

    def handle_action(self):
        logger.debug("Received invalid action from client. Lock: %s", self.lock)
        self.protocol.send(const.RESPONSE_INVALID_ACTION)



import logging
import threading
import time

from tcpnetlock.server.context import Context

logger = logging.getLogger(__name__)


class BackgroundThread(threading.Thread):

    daemon = True
    iteration_wait = 5
    min_age = 5

    def __init__(self, context: Context, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._context = context

    def run(self):
        while True:
            self.cleanup_old_locks()

    def cleanup_old_locks(self):
        logger.info("Cleaning up old locks...")
        for key in self._get_keys():
            try:
                self._check_key(key)
            except:  # noqa: E722 we need to ignore any error
                logger.exception("Exception detected in BackgroundThread while checking key '%s'", key)
        logger.info("Cleaning finished")
        self._wait()

    def _wait(self):
        time.sleep(self.iteration_wait)

    def _get_keys(self):
        # try:
        #     return list(Context.LOCKS.keys())
        # except RuntimeError:
        #     logger.info("RuntimeError detected when trying to get list of keys. "
        #                      "Will retry holding GLOBAL_LOCK", exc_info=True)
        #     with Context.GLOBAL_LOCK:
        #         return list(Context.LOCKS.keys())

        # Always get the GLOBAL_LOCK. It would be ok if we get an exception here, but before using the code above,
        # we must be sure that the iteration done in the code do not prevents the main server modify LOCKS,
        # and that there is NO way that our iteration could generate any inconvenience there.

        with self._context.global_lock:
            return list(self._context.locks.keys())

    def _check_key(self, key):
        lock = self._context.locks[key]
        if lock.locked:
            return
        if lock.age < self.min_age:
            # Here we're evaluating the lock grant age, so, not sure if this makes sense
            return

        ackquired = False
        try:
            ackquired = lock.acquire_non_blocking()
            if not ackquired:
                logger.info("Weird... couldn't get the lock to delete '%s'", key)
                return

            logger.info("Cleaning up key '%s'", key)
            del self._context.locks[key]

        finally:
            if ackquired:
                lock.release()



"""
Tests for `tcpnetlock.cli.tnl_do` package.
"""

import re
import subprocess

import pytest

from tcpnetlock.cli import tnl_do
from tcpnetlock.client.client import LockClient

from .test_utils import ServerThread
from .test_utils import lock_server
from .test_utils import lock_name
from .test_utils import free_tcp_port

assert lock_server
assert lock_name
assert free_tcp_port


class TestRunWithLock:

    def _run(self, lock_name, lock_server, *args) -> subprocess.CompletedProcess:
        port_arg = [arg for arg in args if arg.startswith('--port=')] or ['--port={port}'.format(port=lock_server.port)]
        base_args = [
            'python', '-m', 'tcpnetlock.cli.tnl_do',
            '--host=localhost',
            '--lock-name={lock_name}'.format(lock_name=lock_name),
        ] + port_arg

        full_args = base_args + list(args)

        completed_process = subprocess.run(full_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return completed_process

    def test_cli_vmstat(self, lock_server: ServerThread, lock_name: str):
        completed_process = self._run(lock_name, lock_server, '--', 'vmstat', '1', '1')
        stdout = completed_process.stdout.decode().splitlines()

        # procs -----------memory---------- ---swap-- -----io---- -system-- ------cpu-----
        #  r  b   swpd   free   buff  cache   si   so    bi    bo   in   cs us sy id wa st
        #  1  0      0 484936 1067164 3394552    0    0    54    57  214  537  6  2 92  0  0

        assert completed_process.returncode == 0
        assert len(stdout) == 3
        assert stdout[0].strip().startswith('procs ')
        assert stdout[1].strip().startswith('r ')
        assert re.match(r'^[ 0-9]+$', stdout[2])

    def test_cli_vmstat_with_shell(self, lock_server: ServerThread, lock_name: str):
        completed_process = self._run(lock_name, lock_server, '--shell', '--', 'vmstat 1 1')
        stdout = completed_process.stdout.decode().splitlines()

        # procs -----------memory---------- ---swap-- -----io---- -system-- ------cpu-----
        #  r  b   swpd   free   buff  cache   si   so    bi    bo   in   cs us sy id wa st
        #  1  0      0 484936 1067164 3394552    0    0    54    57  214  537  6  2 92  0  0

        assert completed_process.returncode == 0
        assert len(stdout) == 3
        assert stdout[0].strip().startswith('procs ')
        assert stdout[1].strip().startswith('r ')
        assert re.match(r'^[ 0-9]+$', stdout[2])

    def test_cli_vmstat_with_shell_and_pipe(self, lock_server: ServerThread, lock_name: str):
        completed_process = self._run(lock_name, lock_server, '--shell', '--', 'vmstat 1 1 | grep -v procs')
        #  r  b   swpd   free   buff  cache   si   so    bi    bo   in   cs us sy id wa st
        #  1  0      0 484936 1067164 3394552    0    0    54    57  214  537  6  2 92  0  0
        stdout = completed_process.stdout.decode().splitlines()

        assert completed_process.returncode == 0
        assert len(stdout) == 2
        assert stdout[0].strip().startswith('r ')
        assert re.match(r'^[ 0-9]+$', stdout[1])

    def test_cli_fails_lock_not_granted(self, lock_server: ServerThread, lock_name: str):
        lock_client = lock_server.get_client()
        lock_client.connect()
        assert lock_client.lock(lock_name)

        completed_process = self._run(lock_name, lock_server, '--', 'vmstat', '1', '1')
        assert completed_process.returncode == tnl_do.ERR_LOCK_NOT_GRANTED

    def test_cli_fails_lock_not_granted_with_retry(self, lock_server: ServerThread, lock_name: str):
        lock_client = lock_server.get_client()
        lock_client.connect()
        assert lock_client.lock(lock_name)

        completed_process = self._run(lock_name, lock_server, '--retry=2', '--retry-wait=1', '--', 'vmstat', '1', '1')
        assert completed_process.returncode == tnl_do.ERR_LOCK_NOT_GRANTED

    def test_cli_fails_file_not_found(self, lock_server: ServerThread, lock_name: str):
        completed_process = self._run(lock_name, lock_server, '--', '/non/existing/binary', '1', '1')
        assert completed_process.returncode == tnl_do.ERR_FILE_NOT_FOUND

    def test_cli_fails_with_invalid_options(self, lock_server: ServerThread, lock_name: str):
        completed_process = self._run(lock_name, lock_server, '--shell', '--', 'vmstat', 'other-arg')
        assert completed_process.returncode == tnl_do.ERR_INVALID_OPTIONS

    def test_cli_fails_with_connection_refused(self, lock_server: ServerThread, lock_name: str):
        lock_client = LockClient("localhost", port=lock_server.port + 1)  # assume this is free
        with pytest.raises(ConnectionRefusedError):
            lock_client.connect()
        completed_process = self._run(
            lock_name, lock_server,
            '--port={port}'.format(port=lock_server.port + 1),
            '--', 'vmstat', '1', '1')
        assert completed_process.returncode == tnl_do.ERR_CONNECTION_REFUSED

    def test_cli_fails_with_exit_status_of_command(self, lock_server: ServerThread, lock_name: str):
        completed_process = self._run(lock_name + '1', lock_server, '--shell', 'exit 5')
        assert completed_process.returncode == 5

        completed_process = self._run(lock_name + '2', lock_server, '--shell', 'exit 8')
        assert completed_process.returncode == 8

    def test_cli_uses_environment_variable(self, free_tcp_port):
        args = [
            'timeout', '1s',
            'env', 'TCPNETLOCK_PORT={port}'.format(port=free_tcp_port),
            'python', '-m', 'tcpnetlock.cli.tnl_do',
            '--',
            'vmstat', '1',
        ]
        completed_process = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        assert completed_process.returncode == tnl_do.ERR_CONNECTION_REFUSED



"""
Unittests for `tcpnetlock.server.action` package.
"""

from tcpnetlock.server import action


class TestActionStr:
    """
    Test Action.__str__() do not fails
    """

    def test_action_str(self):
        act = action.Action.from_line('action_name')
        assert act.is_valid()
        str(act)

    def test_action_str_invalid_action(self):
        act = action.Action.from_line(',no:action')
        assert not act.is_valid()
        str(act)

    def test_action_str_with_params(self):
        act = action.Action.from_line('action_name,key1:value1')
        assert act.is_valid()
        str(act)

    def test_action_str_with_params_invalid_action(self):
        act = action.Action.from_line('action_name,:nokey')
        assert not act.is_valid()
        str(act)



import argparse
import logging

logger = logging.getLogger(__name__)


class BaseMain:

    def __init__(self):
        self.parser = argparse.ArgumentParser()
        self.args = None

    def add_logging_arguments(self):
        self.parser.add_argument(
            "--debug",
            default=False,
            action='store_true')

        self.parser.add_argument(
            "--info",
            default=False,
            action='store_true')

    def add_app_arguments(self):
        raise NotImplementedError()

    def create_parser(self):
        self.add_app_arguments()
        self.add_logging_arguments()

    def create_args(self, args):
        assert self.parser
        self.args = self.parser.parse_args(args)

    def setup_logging(self):
        assert self.args
        if self.args.debug:
            logging.basicConfig(level=logging.DEBUG)
        elif self.args.info:
            logging.basicConfig(level=logging.INFO)
        else:
            logging.basicConfig(level=logging.WARNING)

    def run(self, args):
        self.create_parser()
        self.create_args(args)
        self.setup_logging()
        self.main()

    def main(self):
        raise NotImplementedError()


class PositiveInteger:

    def __init__(self, allow_zero=True):
        self.allow_zero = allow_zero

    def __call__(self, value):
        int_value = int(value)
        if self.allow_zero:
            if int_value < 0:
                raise argparse.ArgumentTypeError("{value} is an invalid positive int value".format(value=int_value))
        else:
            if int_value <= 0:
                raise argparse.ArgumentTypeError("{value} is an invalid positive int value".format(value=int_value))

        return int_value


"""
Example:


class Main(utils.BaseMain):

    def add_app_arguments(self):
        parser = self.parser
        parser.add_argument("--host",
                            default='localhost')

        parser.add_argument("--port",
                            default=server.TCPServer.DEFAULT_PORT,
                            type=int)

    def main(self):
        pass

def main():
    Main().run(args=sys.argv[1:])


if __name__ == '__main__':
    main()


"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = []

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="Horacio G. de Oro",
    author_email='hgdeoro@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Networking',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
    description="Network lock based on TCP sockets",
    entry_points={
        'console_scripts': [
            'tcpnetlock_server=tcpnetlock.cli.tnl_server:main',
            'tcpnetlock_client=tcpnetlock.cli.tnl_client:main',
            'tcpnetlock_do=tcpnetlock.cli.tnl_do:main'
        ],
    },
    scripts=[],
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='tcpnetlock',
    name='tcpnetlock',
    packages=find_packages(include=[
        'tcpnetlock',
        'tcpnetlock.client',
        'tcpnetlock.cli',
        'tcpnetlock.server',
    ]),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/hgdeoro/tcpnetlock',
    version='0.1.7',
    zip_safe=False,
)



"""
Tests for `tcpnetlock.cli.tnl_client` package.
"""
import subprocess
import uuid

from tcpnetlock.cli import tnl_client
from tests.test_utils import ServerThread
from .test_utils import BaseTest
from .test_utils import lock_server, free_tcp_port, lock_name

assert lock_server
assert free_tcp_port
assert lock_name


class TestClientCli(BaseTest):

    def test_report_cant_connect(self, free_tcp_port):
        assert free_tcp_port > 0
        args = [
            'python', '-m', 'tcpnetlock.cli.tnl_client',
            '--debug',
            '--port={port}'.format(port=free_tcp_port),
            str(uuid.uuid4())
        ]
        completed_process = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        assert completed_process.returncode == tnl_client.ERR_CONNECTION_REFUSED
        assert completed_process.stderr.decode().find('[Errno 111]') >= 0

    def test_cli_fails_lock_not_granted(self, lock_server: ServerThread, lock_name: str):
        lock_client = lock_server.get_client()
        lock_client.connect()
        assert lock_client.lock(lock_name)

        args = [
            'python', '-m', 'tcpnetlock.cli.tnl_client',
            '--debug',
            '--port={port}'.format(port=lock_server.port),
            lock_name
        ]
        completed_process = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        assert completed_process.returncode == tnl_client.ERR_LOCK_NOT_GRANTED
        assert completed_process.stderr.decode().find('not granted') >= 0

    def test_cli_uses_environment_variable(self, free_tcp_port):
        args = [
            'timeout', '1s',
            'env', 'TCPNETLOCK_PORT={port}'.format(port=free_tcp_port),
            'python', '-m', 'tcpnetlock.cli.tnl_client',  '--debug', 'some-lock'
        ]
        completed_process = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        assert completed_process.returncode == tnl_client.ERR_CONNECTION_REFUSED



import re

RESPONSE_OK = 'ok'
RESPONSE_ERR = 'err'
RESPONSE_INVALID_ACTION = 'bad-action'
RESPONSE_INVALID_REQUEST = 'bad-request'
RESPONSE_LOCK_NOT_GRANTED = 'not-granted'
RESPONSE_RELEASED = 'released'
RESPONSE_SHUTTING_DOWN = 'shutting-down'
RESPONSE_PONG = 'pong'
RESPONSE_STILL_ALIVE = 'alive'

ACTION_LOCK = 'lock'
ACTION_RELEASE = 'release'
ACTION_SERVER_SHUTDOWN = '.server-shutdown'
ACTION_PING = '.ping'
ACTION_KEEPALIVE = '.keepalive'
ACTION_STATS = '.stats'

VALID_LOCK_NAME_RE = re.compile(r'^[a-zA-Z0-9_-]+$')
VALID_CLIENT_ID_RE = re.compile(r'^[a-zA-Z0-9\.@_-]+$')
VALID_CHARS_IN_LOCK_NAME_RE = re.compile(r'[a-zA-Z0-9_-]')

NEW_LINE_BINARY = '\n'.encode()
